#!/usr/bin/perl
# buildfam.pl -- does a great many things, too numerous to list on one line.
# Copyright 2002 Christopher League <league@contrapunctus.net>
# This is free software under the GNU GPL; it comes with NO WARRANTY.

use strict;
use File::Basename;
use File::Path;
use NDBM_File;
use Fcntl;
use POSIX qw(strftime);

## Grab arguments
my $afmdir = shift;
my $tag = shift;

## Switch to working directory
chdir "work" || die "chdir work: $!";

## Index the AFM files, mapping Postscript FontName to filename.
my %afmIndex;
tie (%afmIndex, 'NDBM_File', 'afms', O_RDWR|O_CREAT, 0644);
my $edition = $afmIndex{edition};
&indexAFMdir;

## Keep persistent track of human-readable font names
my %names;
tie (%names, 'NDBM_File', 'names', O_RDWR|O_CREAT, 0644);

## Berry weight codes ('h'), names ('heavy'), and LaTeX codes ('hb').
## See "ttf2tex" manual for extended LaTeX codes.
my %weightMap = 
    ('a' => {nfss=>"ul", name=>"thin"},
     'b' => {nfss=>"b",  name=>"bold"},
     'c' => {nfss=>"cb", name=>"black"},
     'd' => {nfss=>"db", name=>"demi"},
     'h' => {nfss=>"hb", name=>"heavy"},
     'j' => {nfss=>"el", name=>"extralight"},
     'k' => {nfss=>"k",  name=>"book"},
     'l' => {nfss=>"l",  name=>"light"},
     'm' => {nfss=>"mb", name=>"medium"},
     'p' => {nfss=>"pb", name=>"poster"},
     'r' => {nfss=>"m",  name=>"regular"},
     's' => {nfss=>"sb", name=>"semibold"},
     'u' => {nfss=>"ub", name=>"ultra"},
     'x' => {nfss=>"eb", name=>"extrabold"},
     );

## Same for widths
my %widthMap =
    ('c' => {nfss=>"c",  name=>"condensed"},
     'e' => {nfss=>"x",  name=>"expanded"},
     'n' => {nfss=>"sc", name=>"narrow"},
     'o' => {nfss=>"uc", name=>"ultracondensed"},
#     'p' => {nfss=>"", name=>"compressed"},
     'q' => {nfss=>"ec", name=>"extracompressed"},
     'r' => {nfss=>"m", name=>"normal"},
#     't' => {nfss=>"", name=>"thin"},
#     'u' => {nfss=>"", name=>"ultracompressed"},
     'v' => {nfss=>"ex", name=>"extraexpanded"},
     'w' => {nfss=>"ux", name=>"wide"},
     'x' => {nfss=>"sx", name=>"extended"},
     );

## Encoding information:
##   'etx'  is the .etx ised by fontinst to reencode.
##   'enc'  is the PostScript encoding vector used by dvips.
##   'ps'   is the PostScript name associated with the encoding vector.
##   'nfss' is the name used internally in NFSS.
my %encodingMap =
    ('7t' => {etx=>"fs7t", nfss=>"OT1"},
     '8c' => {etx=>"ts1", nfss=>"TS1"},
     '8t' => {etx=>"t1", nfss=>"T1" },
     '8y' => {etx=>"8y", nfss=>"LY1", enc=>"texnansi", ps=>"TeXnANSIEncoding"},
     '8r' => {etx=>"8r", enc=>"8r", ps=>"TeXBase1Encoding"},
     'fs8r' => {etx=>"fs8r", enc=>"fs8r", ps=>"FontSiteTeXEncoding"},
     'fs8xa' => {etx=>"fs8xa", enc=>"fs8xa", ps=>"FontSiteExpertEncodingA"},
     'fs8xb' => {etx=>"fs8xb", enc=>"fs8xb", ps=>"FontSiteExpertEncodingB"},
     );

## Automatic NFSS substitutions
##   =>  means a noisy substitution (warning is issued)
##   ->  means a silent substitution
my @shapeSubs = ("it=>sl", "sl=>it");
my @seriesSubs = ("bx->b", "bx->db");

## LaTeX version string
my $versionString = (strftime "%Y/%m/%d", gmtime) . 
    "%fs500tex-".`cat ../VERSION`."%cd".$edition.
    "%".`whoami`."@".`hostname -s`;
$versionString =~ s/\s//g;
$versionString =~ s/%/ /g;

## Read the family control file
my %family;
&parseControlFile (\%family, "../faces/$tag");
## Check for required fields
for my $field ('name', 'fonts', 'fn') {
    exists $family{$field} or die "$tag did not specify '$field'";
}
my $name = $family{name};
my $supplier = "5";
my $pre = "$supplier$family{fn}";
$names{$tag} = $name;

my %fonts;
&parseFontsField (\%fonts, $family{fonts});

my %tweaks;
&parseTweaks (\%tweaks, $family{tweaks});

## Create non-TeX maps
&gsmap;
&x11map;

## Open a pipe to fontinst..
$ENV{TEXINPUTS} = "../tools/::";
open FONTINST, "|tex >/dev/null" or die "cannot run tex: $!";
print FONTINST "\\input fontinst.sty\n";
print FONTINST "\\installfonts\n";

my (%metrics, %nfss, %nfssv);
&convert;
&slant if $family{slant};
&reencode;
&type1maps;

if ($family{class} =~ /symbol/) {
    &symbolFontDef;
}
else {
    &latexpkg;
## Create a family for each target encoding.  It is convenient to do
## OT1 (7t) last, so then the %nfss table is set up for writing the
## documentation in terms of OT1.
    foreach my $enc ('8y', '8t', '8c', '7t') {
        combineFamily ($enc, 1) if $family{osf};
        combineFamily ($enc, 0);
    }
    &expertFontDef($pre."x") if $family{expert};
    &expertFontDef($pre."jx") if $family{expert} and $family{osf};
}

print FONTINST "\\endinstallfonts\\bye\n";
close FONTINST;

## Generate documentation
&prepareGlyphChart;
&listSnippet;
&docSnippet;
&docIndices;

## Clean up and shut down.
&compileFonts;
#exit 0;
&cleanUp;
untie %names;
untie %afmIndex;
exit 0;


## Index the AFM files, mapping Postscript FontName to filename.
## Also record the copyright notice here, just because it is convenient.
sub indexAFMdir
{
    if ($afmIndex{dir} ne $afmdir)
    {
        # Guess which CD we've got.
        my ($afmd) = glob("$afmdir/");
        if (-d "$afmd/Collner Grotesk") {
            $edition = 3;       # 2003
        } elsif (-f "$afmd/colgro__.pfb" or -f "$afmd/COLGRO__.PFB") {
            $edition = 1;       # 2001
        } elsif (-f "$afmd/berlg___.pfb" or -f "$afmd/BERLG___.PFB") {
            $edition = 0;       # 2000
        } else {
            die "Cannot guess your CD version";
        }
        print "Looks like version $edition...\n";
        print "Indexing $afmdir...\n";
        $afmIndex{edition} = $edition;
        $afmIndex{dir} = $afmdir;
        system("echo $edition >EDITION");

        ## look for *.afm (case insensitive)
        my @afmFiles = (glob("$afmdir/*/*.[Aa][Ff][Mm]"),
                        glob("$afmdir/*.[Aa][Ff][Mm]"));
        print "    " . ($#afmFiles+1) . " files\n";

        ## open each AFM file to find FontName attribute
        my $n = 0;
        while (my $afm = shift @afmFiles) {
            open(AFM, "$afm") or die "Cannot open $afm: $!";
            my $psname = "";
            my $notice = "";
          line:
            while (<AFM>) {
                if (/^FontName\s+(\S+)\s*$/) {
                    $psname = $1;
                } elsif (/^Notice\s+([^\r]*)[\r\s]*$/) {
                    $notice = $1;
                } elsif (/^StartCharMetrics/) {
                    last line;
                }
            }
            if ($psname) {
                $afmIndex{$psname} = basename($afm) . "#$notice";
                $n++;
            }
            close(AFM);
        }
        print "    $n fonts\n";
    }
}

## Grab one-letter codes and names from a Fontname map file.
sub fromFontnameMap
{
    my ($map, $file) = @_;
    open (FM, $file) or die "cannot open $file: $!";
  line:
    while (<FM>) {
        next line if /^\@c/;
        $map->{$1} = lc $2 if /^([a-z])\s+(\S+)/;
    }
    close(FM);
}

## The control file is just a series of "key = value" lines.
## Multi-line values are supported as follows:
##   key = <<
##   line 1 here
##   line 2 here
##   <<

sub parseControlFile {
    my ($family, $path) = @_;
    open (FAM, $path) or die "$path: $!";
  ENTRY:
    while (<FAM>) {
        chop;
        if (/^\s*\#/) { }
        elsif (/^\s*(\S+)\s*=\s*<</) {
            my $key = $1;
            my $val = "";
            while(<FAM>) {
                if (/^<</) { $family->{$key} = $val; next ENTRY; }
                else { $val .= $_; }
            }
            last ENTRY;
        }
        elsif (/^\s*(\S+)\s*=\s*(.*)$/) {
            $family->{$1} = $2;
        }
    }
    close(FAM);
}

sub parseFontsField {
    my ($map, $string) = @_;
    foreach my $line (split /\n/, $string) {
        if ($line =~ /(\S+)\s*(\S+)/) {
            my $code = $1;
            my $name_regex = $2;
            ## First check whether name matches exactly
            if (!defined ($afmIndex{$name_regex})) {
                my @matches = grep /$name_regex/, (keys %afmIndex);
                scalar @matches > 0 or die "$name_regex did not match as regex\n";
                scalar @matches < 2 or die "$name_regex is ambiguous\n";
                $name_regex = $matches[0];
            }
            $map->{$code} = $name_regex;
        }
    }
}

sub parseTweaks {
    my ($map, $string) = @_;
    foreach my $tw (split /\s+/, $string) {
        my ($fd, $mtx) = split ':', $tw;
        $map->{$fd} = $mtx;
    }
}

## Ghostscript uses a file which maps Postscript font names to the 
## PFB files that define them.  The format is:
##    /PostscriptName  (FILENAME.PFB) ;

sub gsmap {
    open (XM, ">$pre.gs") or die "cannot write $pre.gs: $!";
    foreach my $name (values %fonts) {
        my $pfb = canonicalPFB($name);
        print XM "/$name ($pfb) ;\n";
    }
    close(XM);
}

## X11 uses a file which relates PFB files with descriptor strings:
## FILENAME.PFB -supplier-family-weight-slant-width--0-0-0-0-p-0-reg-enc
## Technically, the field between the two dashes (after width) is for
## "additional styles", which you'd think would be a good place to
## include small caps, expert sets, etc.  But unfortunately, the GTK
## font selection dialog ignores this field!

sub x11map {
    ## We build the map in this hash first, then write it to a file in
    ## a post-pass.  This way we can tell if we somehow generate
    ## non-unique X11 descriptors.
    my %map;

    while (my ($desc, $name) = each %fonts) {
        # name of PFB file
        my $pfb = canonicalPFB($name);

        # dissect the font descriptor
        my ($enc, $vs, $wt, $wd) = decodeFontDescriptor($desc);
        my $sl = "r";;
        if ($vs =~ /i/) { $sl = "i"; }
        elsif ($vs =~ /o/) { $sl = "o"; }

        # family name -- may be augmented by "Expert" or "SC+OSF"
        my $xfamily = $family{name};
        if ($enc eq "8x")     { $xfamily .= " Expert"; }
        elsif ($vs =~ /c.*j/) { $xfamily .= " SC OSF"; }
        elsif ($vs =~ /c/)    { $xfamily .= " SC"; }
        elsif ($vs =~ /j/)    { $xfamily .= " OSF"; }
        elsif ($vs =~ /g/)    { $xfamily .= " Small"; }
        elsif ($vs =~ /d/)    { $xfamily .= " Display"; }
        elsif ($vs =~ /l/)    { $xfamily .= " Open"; }
        elsif ($vs =~ /a/)    { $xfamily .= " Alternate"; }

        # encoding
        my $xenc;
        if ($enc eq "8a" and !($vs =~ /x/)) { $xenc = "iso8859-1"; } 
        else { $xenc = "adobe-fontspecific"; }

        # construct the entry
        my $x = ("-fontsite-$xfamily-$weightMap{$wt}{name}-$sl-" .
                 "$widthMap{$wd}{name}--0-0-0-0-p-0-$xenc");

        if (defined($map{$x})) {
            die "$pfb and $map{$x} yielded the same X descriptor: $x";
        }
        $map{$x} = $pfb;
    }

    open (XM, ">$pre.x11") or die "cannot write $pre.x11: $!";
    while (my ($x, $pfb) = each %map) {
        print XM "$pfb $x\n";
    }
    close(XM);
}

###   - Nomenclature  

sub decodeFontDescriptor {
    my $desc = shift;
    my ($enc, $vs, $wt, $wd);
    $desc =~ /^(.)([^5-9]*)(([5-9].)(.)?)?/
        or die "bad descriptor '$desc'";
    $wt = $1;
    $vs = $2;
    $enc = $4 || "8a";
    $wd = $5 || "r";
    # undo shortening tricks
    $vs =~ s/2/cj/;             # 2 == cj  (SC&OsF)
    if ($enc =~ /^9([cety])/) {
        if ($1 eq "e") { $enc = "8t"; }
        elsif ($1 eq "t") { $enc = "7t"; }
        else { $enc = "8$1"; }
    }
    return ($enc, $vs, $wt, $wd);
}

sub encodeFontDescriptor {
    my ($enc, $var, $wt, $wd) = @_;
    $var = join("", sort (split("", $var)));
    my $desc = "$wt$var";
    if ($wd ne "r") {
        $desc .= "$enc$wd";
    } elsif ($enc ne "8a") {
        $desc .= "$enc";
    }
    # tricks to shorten descriptors
    if (length $desc > 5) {
        # variants "cj" => "2"
        if ($var =~ /c/ and $var =~ /j/) {
            $var =~ s/c//; $var =~ s/j//;
            return encodeFontDescriptor($enc, "2$var", $wt, $wd);
        }
        # variant "x" merged into encoding
        if ($var =~ /x/) {
            $var =~ s/x//;
            if ($enc eq "8t") { $enc = "9e"; }
            elsif ($enc eq "7t") { $enc = "9t"; }
            elsif ($enc =~ /^8([cy])$/) { $enc = "9$1"; }
            else { die "unexpected descriptor problem"; }
            return encodeFontDescriptor($enc, $var, $wt, $wd);
        }
    }
    return $desc;
}

sub canonicalPFB {
    my ($name) = @_;
    my ($pfb, $notice) = split '#', $afmIndex{$name};
    $pfb = lc $pfb;
    $pfb =~ s/\.afm$/.pfb/;
    return $pfb;
}

sub descriptorToNFSS {
    my ($var, $wt, $wd) = @_;
    my $series = $weightMap{$wt}{nfss};
    $series .= $widthMap{$wd}{nfss} unless $wd eq "r";

    my $shape = "";
    if ($var =~ /c/) { $shape .= "sc"; }
    if ($var =~ /d/) { $shape .= "dp"; }  # display
    if ($var =~ /l/) { $shape .= "op"; }  # open
    if ($var =~ /g/) { $shape .= "sm"; }  # SmallText 
    if ($var =~ /a/) { $shape .= "alt"; }  # Alternative
    if ($var =~ /i/) { $shape .= "it"; }
    if ($var =~ /o/) { $shape .= "sl"; }
    if (!$shape) { $shape = "n"; }
    
    return ($series, $shape);
}

sub descriptorToEnglish {
    my ($fd) = @_;
    my ($en, $vs, $wt, $wd) = decodeFontDescriptor($fd);
    my $text = "";
    $text .= $weightMap{$wt}{name} unless $wt eq 'r';
    if ($vs =~ /i/) { $text .= " Italic"; }
    elsif ($vs =~ /o/) { $text .= " Oblique"; }
    if ($vs =~ /c.*j/) { $text .= " SC OsF"; }
    elsif ($vs =~ /c/) { $text .= " SC"; }
    elsif ($vs =~ /j/) { $text .= " OsF"; }
    elsif ($vs =~ /g/) { $text .= " Small"; }
    elsif ($vs =~ /d/) { $text .= " Display"; }
    elsif ($vs =~ /l/) { $text .= " Open"; }
    elsif ($vs =~ /a/) { $text .= " Alternate"; }
    $text .= " ".$widthMap{$wd}{name} unless $wd eq 'r';
    $text =~ s/\b(\w)/uc($1)/eg;
    $text =~ s/^\s//g;
    return $text;
}

# Convert all outlines to mtx+pl
sub convert {
    while (my ($fd, $psName) = each %fonts) {
        my $pfb = canonicalPFB($psName);
        my ($afm, $notice) = split '#', $afmIndex{$psName};
        mylink ($afm, "$pre$fd.afm");
        my ($enc, $vs, $wt, $wd) = decodeFontDescriptor($fd);
        fixExpert ($fd) if $vs =~ /x/;
        print FONTINST "\\fromafm{$pre$fd}";
        $metrics{$fd}{ps} = $psName;
        $metrics{$fd}{pfb} = $pfb;
        $family{osf} = 1 if ($vs =~ /j/);
        $family{expert} = 1 if ($vs =~ /x/);
        $family{notice} = $notice if $notice;
    }
}

# Artificially slant fonts, if requested
sub slant {
    my $radians = $family{slant} * (3.14159 / 180);
    my $tangent = abs (sin ($radians) / cos ($radians));
    my $angle = int($tangent * 1000);

    foreach my $fd (keys %metrics) {
        my ($en, $vs, $wt, $wd) = decodeFontDescriptor($fd);
        unless ($vs =~ /[io]/) {
            my $sfd = encodeFontDescriptor($en, "o$vs", $wt, $wd);
            print FONTINST "\\transformfont{$pre$sfd}{\\slantfont{$angle}{\\frommtx{$pre$fd}}}";
            $metrics{$sfd} = deep_copy($metrics{$fd});
            $metrics{$sfd}{slant} = $angle;
        }
    }
}

sub mylink {
    my ($oldBase, $newName) = @_;
    my ($oldPath) = glob("$afmdir/*/$oldBase");
    (unlink($newName) or die "could not unlink $newName")
        if (-f $newName);
    (symlink ($oldPath, $newName) 
     or die "could not symlink $oldPath: $!");
}

# Reencode AdobeStandard fonts to TeXBase1
sub reencode {
    foreach my $fd (keys %metrics) {
        my ($enc, $vs, $wt, $wd) = decodeFontDescriptor($fd);
        if ($enc eq "8a") {
            my $nfd;
            # At this point, if a font has the variant "x", that 
            # means it is really a "raw" expert set.
            if ($vs =~ /x/) {
                $vs =~ s/x//;
                $nfd = encodeFontDescriptor("8x", $vs, $wt, $wd);
                $metrics{$nfd} = deep_copy($metrics{$fd});
                my $tr = "\\frommtx{$pre$fd}";
                unless ($metrics{$fd}{expertokay}) {
                    $metrics{$nfd}{enc} = ($edition >= 3)? "fs8xb" : "fs8xa";
                    $tr = "\\reencodefont{$metrics{$nfd}{enc}}{$tr}";
                }
                print FONTINST "\\transformfont{$pre$nfd}{$tr}";
            } else {
                $nfd = encodeFontDescriptor("8r", $vs, $wt, $wd);
                $metrics{$nfd} = deep_copy($metrics{$fd});
#                print "\\transformfont{$pre$nfd}{"
#                    ."\\reencodefont{fs8r}{\\frommtx{$pre$fd}}}\n";
                print FONTINST "\\transformfont{$pre$nfd}{"
                    ."\\reencodefont{fs8r}{\\frommtx{$pre$fd}}}";
                $metrics{$nfd}{enc}   = "fs8r";
            }
            # From now on, forget about the un-re-encoded version
            delete $metrics{$fd};
        }
    }
}

# Create the dvips and pdftex maps
sub type1maps {
    open (DVI, ">$pre.map") or die "cannot write $pre.map: $!";
  font:
    foreach my $fd (keys %metrics) {
        my ($enc, $vs, $wt, $wd) = decodeFontDescriptor($fd);
        next font if $enc eq "8a";
        my $tr = "";
        my $incl = "<" . $metrics{$fd}{pfb};
        my $enc = $metrics{$fd}{enc};
        if ($enc) {
            $incl = "<".$encodingMap{$enc}{enc}.".enc " . $incl;
            $tr .= " ".$encodingMap{$enc}{ps}." ReEncodeFont";
        }
        if ($metrics{$fd}{slant}) {
            $tr .= " .$metrics{$fd}{slant} SlantFont";
        }
        printf DVI "%-9s %-20s \"%s\" %s\n", 
        ("$pre$fd", $metrics{$fd}{ps}, $tr, $incl);
    }
    close(DVI);
}

# Create the LaTeX package file
sub latexpkg {
    open (STY, ">$tag.sty") or die "cannot write $tag.sty: $!";
    my $isrm = "true";
    my $dflt = "rm";
    if ($family{class} =~ /sans/) { 
        $isrm = "false"; 
        $dflt = "sf";
    }
    my $osfexec = "\\PackageWarningNoLine{$tag}{Old-style figures not supported}";
    if ($family{class} =~ /osf/) {
        $osfexec = "\\def\\${tag}prefix{${pre}j}";
    }
    print STY
"\\NeedsTeXFormat{LaTeX2e}
\\ProvidesPackage{$tag}[$versionString]
\\newif\\if${tag}rm\\${tag}rm$isrm
\\def\\${tag}prefix{$pre}
\\DeclareOption{rm}{\\${tag}rmtrue}
\\DeclareOption{sf}{\\${tag}rmfalse}
\\DeclareOption{osf}{$osfexec}
\\ExecuteOptions{$dflt}
\\ProcessOptions\\relax
\\if${tag}rm
\\renewcommand{\\rmdefault}{\\${tag}prefix}
\\else
\\renewcommand{\\sfdefault}{\\${tag}prefix}
\\fi
\\endinput
";
    close(STY);
}

sub combineFamily {
    my ($enc, $osf) = @_;
    my $nenc = $encodingMap{$enc}{nfss};
    my $nfam = $pre;  
    $nfam .= "j" if $osf;
    %nfssv = ();
    my $fdpath = lc $nenc . $nfam . ".fd";
    open FD, ">$fdpath" or die "cannot write $fdpath: $!";
    print FD "\\ProvidesFile{$fdpath}[$versionString]\n";
    print FD "\\DeclareFontFamily{$nenc}{$nfam}{}\n";

    foreach my $fd (keys %metrics) {
        my ($en, $vs, $wt, $wd) = decodeFontDescriptor($fd);
        if ($en eq "8r") {
            my $files = "$pre$fd";
            my $nvs = $vs;
            # was a tweak specified?
            if (exists $tweaks{$fd}) {
                $files .= ",$tweaks{$fd}";
            }
            # is there a matching expert set available?
            my $efd = encodeFontDescriptor("8x", $vs, $wt, $wd);
            if (exists $metrics{$efd}) {
                $files .= ",$pre$efd";
                $nvs .= 'x';
            }
            # is there a matching sc/osf available?
            if (! ($vs =~ /j/)) {
                my $jfd = encodeFontDescriptor($en, "j$vs", $wt, $wd);
                unless (exists $metrics{$jfd}) {
                    $jfd = encodeFontDescriptor($en, "cj$vs", $wt, $wd);
                }
                if (exists $metrics{$jfd}) {
                    # maybe merge it in
                    $files .= ",unsetnum,$pre$jfd"
                        if $osf or $enc eq "8c";
                    $nvs .= 'j'
                        if $osf and ! ($enc eq "8c");
                }
            }
            $files .= ",resetosf,fakets1" if $enc eq "8c";
            $files .= ",fixfs";
            my ($se, $sh) = descriptorToNFSS($vs, $wt, $wd);
            my $nfd = encodeFontDescriptor($enc, $nvs, $wt, $wd);
            $nfssv{"$se:$sh"} = 1;
            if ($osf and !($nvs =~ /j/)) {
               print FD "\\DeclareFontShape{$nenc}{$nfam}{$se}{$sh}{<->ssub * $pre/$se/$sh}{}\n";
            } else {
#                print "\\installfont{$pre$nfd}{$files}{$encodingMap{$enc}{etx}}{}{}{}{}{}\n";
                print FONTINST "\\installfont{$pre$nfd}{$files}{$encodingMap{$enc}{etx}}{}{}{}{}{}";
                print FD "\\DeclareFontShape{$nenc}{$nfam}{$se}{$sh}{<->$pre$nfd}{}\n";
                $nfss{"$se:$sh"} = $nfd;
            }
        }
    }
    declareSubstitutions($nenc,$nfam);
    print FD "\\endinput\n";
    close(FD);
}

sub expertFontDef {
    my ($nfam) = @_;
    my $nenc = "U";
    my $fdpath = "u" . $nfam . ".fd";
    %nfssv = ();
    open FD, ">$fdpath" or die "cannot write $fdpath: $!";
    print FD "\\ProvidesFile{$fdpath}[$versionString]\n";
    print FD "\\DeclareFontFamily{$nenc}{$nfam}{}\n";
    foreach my $fd (keys %metrics) {
        my ($en, $vs, $wt, $wd) = decodeFontDescriptor($fd);
        if ($en eq "8x") {
            my ($se, $sh) = descriptorToNFSS($vs, $wt, $wd);
            print FD "\\DeclareFontShape{$nenc}{$nfam}{$se}{$sh}{<->$pre$fd}{}\n";
            $nfssv{"$se:$sh"} = 1;
        }
    }
    declareSubstitutions($nenc,$nfam);
    print FD "\\endinput\n";
    close(FD);
}

## Automatically declare substitutions.
## Keep track of all series/shape combos already
## defined in %nfssv.
sub declareSubstitutions {
    my ($nenc, $nfam) = @_;

    # Shapes first
    foreach my $spec (@shapeSubs) {
        $spec =~ /([a-z]+)([-=])>([a-z]+)/ 
            or die "bad substitution: $spec";
        my $new = $1;  my $tgt = $3;
        my $cmd = ($2 eq "-"? "ssub" : "sub");
        foreach my $ss (keys %nfssv) {
            my ($se, $sh) = split /:/, $ss;
            if ($sh eq $tgt and ! exists $nfssv{"$se:$new"}) {
                print FD "\\DeclareFontShape{$nenc}{$nfam}{$se}{$new}{<->$cmd * $nfam/$se/$sh}{}\n";
                $nfssv{"$se:$new"} = $ss;
            }
        }
    }

    # then Series
    foreach my $spec (@seriesSubs) {
        $spec =~ /([a-z]+)([-=])>([a-z]+)/ 
            or die "bad substitution: $spec";
        my $new = $1;  my $tgt = $3;
        my $cmd = ($2 eq "-"? "ssub" : "sub");
        foreach my $ss (keys %nfssv) {
            my ($se, $sh) = split /:/, $ss;
            if ($se eq $tgt and ! exists $nfssv{"$new:$sh"}) {
                print FD "\\DeclareFontShape{$nenc}{$nfam}{$new}{$sh}{<->$cmd * $nfam/$se/$sh}{}\n";
                $nfssv{"$new:$sh"} = $ss;
            }
        }
    }

}

sub symbolFontDef {
    my $fdpath = "u$pre.fd";
    open FD, ">$fdpath" or die "cannot write $fdpath: $!";
    print FD "\\ProvidesFile{$fdpath}[$versionString]\n";
    print FD "\\DeclareFontFamily{U}{$pre}{}\n";
    foreach my $fd (keys %metrics) {
        my ($en, $vs, $wt, $wd) = decodeFontDescriptor($fd);
        my ($se, $sh) = descriptorToNFSS($vs, $wt, $wd);
        print FD "\\DeclareFontShape{U}{$pre}{$se}{$sh}{<->$pre$fd}{}\n";
        $nfss{"$se:$sh"} = $fd;
    }
    print FD "\\endinput\n";
    close(FD);
}

# The "chart" field gives ranges of characters (in hex)
# to place in a glyph chart, for symbol fonts.
# This routine interprets the ranges.
sub prepareGlyphChart {
    return unless $family{chart};
    my @ranges = split(/[,;\s]+/, $family{chart});
    my %chart;
    foreach my $r (@ranges) {
        if ($r =~ /([^-]+)-([^-]+)/) {
            my $a = hex $1;
            my $b = hex $2;
            die "$tag: invalid range $1->$2" unless $a < $b;
            for (my $i = $a;  $i <= $b;  $i++) {
                $chart{$i} = 1;
            }
        }
        else {
            my $i = hex $r;
            $chart{$i} = 1;
        }
    }
    $family{chart} = \%chart;
}

# Output snippet for font list
sub listSnippet {
    foreach my $c (split /\s+/, $family{class}) {
        open LIST, ">$c.lst" or die "cannot write $c.lst: $!";
        print LIST "\\nextfamily{$tag}\n";
        foreach my $s (sort nfssCmp (keys %nfss)) {
            my $fd = $nfss{$s};
            next if $family{slant} and $fd =~ /o/;
            my $text = descriptorToEnglish($fd);
            if ($family{class} =~ /symbol/) {
                print LIST "\\dosymbolfont{$pre$fd}{$family{name}}\n";
            } else {
                print LIST "\\dofont{$pre$fd}{$family{name} $text}\n";
            }
        }
        close(LIST);
    }
}

# Output snippet for documentation
sub docSnippet {
    my $synopsis;
    if ($family{class} =~ /symbol/) {
        $synopsis = "\\usepackage\@{pifont\@}  % \\Pifont\@{$pre\@}";
    } else {
        ## decide what options go in the \usepackage 
        my @options = ();
        if ($family{osf}) {
            push @options, "osf";
        }
        if ($family{class} =~ /sans/) {
            push @options, "rm";
        }
        else {
            push @options, "sf";
        }
        my $options = join ',', @options;
        $synopsis = "\\usepackage[$options]\@{$tag\@}  % $pre";
    }

    ## fix up the copyright notice
    $family{notice} =~ 
        s/(Copyright|)\s+?([0-9]+)/$family{name} is copyright \@copyright\{\}$2 by/;
    $family{notice} =~ s/www.fontsite.com//;
    $family{notice} =~ s/\. \b/.  /g;   # double-space after period
    $family{blurb} .= $family{notice};

    ## indicate whether we cheated
    $family{blurb} .= "  The slanted form is artificial."
        if $family{slant};
    $family{demo} = "\@texfontdemo" unless $family{demo};

    ## cross-references by classification
    $family{blurb} .= "  [$family{class}]";

    ## write the snippet
    open DOC, ">$pre.tex" or die "cannot write $pre.tex: $!";
    print DOC <<END;
\@unnumberedsec $family{name}
\@cindex $family{name} \@code{$family{fn}}
\@cindex \@code{$family{fn}} ($family{name})
\@example
$synopsis
\@end example
$family{blurb}
\@multitable \@columnfractions .1 .08 .82
END
    foreach my $s (sort nfssCmp (keys %nfss)) {
        my $fd = $nfss{$s};
        print DOC <<END;
\@item \@code{$fd} \@tab $s \@tab $family{demo}\{$pre$fd\}
END
    }
    print DOC "\@end multitable\n";

    ## do we need a glyph chart?
    if ($family{chart} and $family{fonts} =~ /(\S+)/) {
        my $fd = $1;
        my %chart = %{ $family{chart} };
        print DOC "\@iftex\n";
        print DOC "\@font\@demofont=$pre$1 scaled \@magstep5\n";
        print DOC "{\@demofont\@noindent\n";
        foreach my $i (sort {$a <=> $b} (keys %chart)) {
            # see whether this character is printable
            my $char = sprintf "%c", $i;
            $char = " " unless $char =~ /[[:print:]]/;
            $char = " " if $char =~ /[\@\{\}\\\~\>\|\"\<]/;
            printf DOC 
                "\@hbox to 3cm{\@t{\"%x $char}\@hskip 6pt\@char$i\@hfill} ",
                $i;
        }
        print DOC "}\n\@end iftex\n";
    }
    close(DOC);
}

# Generate samples for each classification
sub docIndices {
    unlink <*.idx>;
    my @nfkeys = sort nfssCmp (keys %nfss);

    # the extended classification is unique because we need to 
    # show several different fonts -- one for each series.
    if ($family{class} =~ s/\bextended\b//) {
        my %series;
        my $m = "";
        foreach my $s (@nfkeys) {
            my ($se,$sh) = split ':', $s;
            unless (exists $series{$se}) {
                $m .= "\@extglyphdemo"."{$pre$nfss{$s}}";
                $series{$se} = $nfss{$s};
            }
        }
        writeIdx("extended", $m);
    }

    # the symbol classification is special too.
    # select 6 or so characters from the map.
    if ($family{chart} and $family{class} =~ s/\bsymbol\b//
        and $family{fonts} =~ /(\S+)/) {
        my $fd = $1;
        my @is = keys %{ $family{chart} };
        my $entry = "\@infont{$pre$fd, ";
        my $n = 0;
      LOOP:
        for (my $i = 0;  $n < 7 && $i <= $#is;  $i++) {
            my $char = sprintf "%c", $is[$i];
            next LOOP unless $char =~ /[[:print:]]/;
            next LOOP if $char =~ /[\@\{\}\\\~\>\|\"\<]/;
            $entry .= $char;
            $n++;
        }
        writeIdx("symbol", "$entry}");
    }

    # For the sc, osf, and expert classifications, we need to
    # look for specific fonts.
    foreach my $s (@nfkeys) {
        my $fd = $nfss{$s};
        my ($enc, $vs, $wt, $wd) = decodeFontDescriptor($fd);
        if ($vs =~ /c/ and $family{class} =~ s/\bsc\b//) {
            writeIdx("sc", "\@texaademo{$pre$fd}");
        }
        if ($vs =~ /j/ and $family{class} =~ s/\bosf\b//) {
            writeIdx("osf", "\@texosfdemo{$pre$fd}");
        }
        if ($vs =~ /x/ and $family{class} =~ s/\bexpert\b//) {
            my $nvs = $vs;
            $nvs =~ s/x//;
            my $nfd = encodeFontDescriptor("8x", $nvs, $wt, $wd);
            writeIdx("expert", "\@texpertdemo{$pre$nfd}");
        }
    }

    # For all remaining classifications, just use the first font.
    my $k = shift @nfkeys;
    my $fd = $nfss{$k};
    foreach my $c (split /\s+/, $family{class}) {
        writeIdx($c, "\@texaademo{$pre$fd}");
    }
}

sub writeIdx {
    my ($c, $m) = @_;
    open(IDX, ">$c.idx") or die "cannot write $c.idx: $!";
    print IDX "\@item \@ref{$family{name}} \@tab $m\n";
    close(IDX);
}


# Sorting NFSS series/shape combinations.  Sort by series
# first, where "m" is the minimum, then by shape, where "n"
# is the minimum.
sub nfssCmp ($$) {
    my ($a, $b) = @_;
    my ($sea, $sha) = split /:/, $a;
    my ($seb, $shb) = split /:/, $b;
    if ($sea eq $seb) {
        if ($sha eq $shb) { return 0; }
        elsif ($sha eq "n") { return -1; }
        elsif ($shb eq "n") { return 1; }
        else { return $sha cmp $shb; }
    }
    elsif ($sea =~ /^m/ and $seb =~ /^m/) { return $sea cmp $seb; }
    elsif ($sea =~ /^m/) { return -1; }
    elsif ($seb =~ /^m/) { return 1; }
    else { return $sea cmp $seb; }
}

sub compileFonts {
    opendir(DIR, ".");
    foreach $_ (readdir(DIR)) {
        if (/^$pre(.*)\.(v?pl)$/) {
            my $ext = $2;
            my ($enc, $vs, $wt, $wd) = decodeFontDescriptor($1);
            unless ($enc eq "8a") {
                if ($ext eq "pl") { system("pltotf $_ >>fontinst.log"); }
                elsif ($ext eq "vpl") { system("vptovf $_ >>fontinst.log"); }
            }
        }
    }
    closedir(DIR);
}

sub cleanUp {
    rmtree($tag);
    mkdir($tag);
    opendir(DIR, ".");
    foreach $_ (readdir(DIR)) {
        if (/$tag\.sty$/ or /\.(idx|lst)$/ or
            /$pre.*\.(tfm|vf|fd|gs|x11|map|tex)$/) 
        { 
            rename $_, "$tag/$_"; 
        } elsif (/^$pre/) { unlink $_; }
    }
    closedir(DIR);
}

# This routine rewrites an AFM file, changing the glyph names.
# The FontSite expert sets have improperly named glyphs in the
# AFM files.

sub fixExpert {
    my $tmp = "tmp$$.afm";
    my $fd = shift;
    my $afm = "$pre$fd.afm";

    my %expertVec;

    if ($edition >= 3) {
        %expertVec = 
            (0x21 => 'oneinferior',
             0x23 => 'threeinferior',
             0x24 => 'fourinferior',
             0x25 => 'fiveinferior',
             0x26 => 'seveninferior',
             0x28 => 'nineinferior',
             0x29 => 'zeroinferior',
             0x2A => 'eightinferior',
             0x2C => 'commasuperior',
             0x2D => 'hyphensuperior',
             0x2E => 'periodsuperior',
             0x2F => 'solidus',  # fractionslash
             0x30 => 'zerosuperior',
             0x31 => 'onesuperior',
             0x32 => 'twosuperior',
             0x33 => 'threesuperior',
             0x34 => 'foursuperior',
             0x35 => 'fivesuperior',
             0x36 => 'sixsuperior', 
             0x37 => 'sevensuperior',
             0x38 => 'eightsuperior',
             0x39 => 'ninesuperior',
             0x3A => 'periodcentered', #?
             0x3B => 'twoperiods',
             0x3D => 'esuperior',
             0x3F => 'currency',
             0x40 => 'twoinferior',
             0x41 => 'dollarsuperior',
             0x42 => 'centsuperior',
             0x43 => 'centoldstyle',
             0x44 => 'dollaroldstyle',
             0x45 => 'euro',
             0x46 => 'florin',
             0x47 => 'Lslash',
             0x48 => 'lslash',
             0x49 => 'dotlessi',
             0x4A => 'fj',
             0x4B => 'kelvin',
             0x4C => 'sterling',
             0x4D => 'pbar',  #?
             0x4E => 'Eng',
             0x4F => 'eng',
             0x50 => 'pesetas',
             0x51 => 'cdblslash', #?
             0x52 => 'rupee', #?
             0x53 => 'sterling',  #again
             0x54 => 'Thorn',
             0x55 => 'thorn',
             0x56 => 'Dbar',   #?
             0x57 => 'eth',
             0x58 => 'times',
             0x59 => 'yen',
             0x5A => 'numero',
             0x5B => 'parenleftsuperior',
             0x5D => 'parenrightsuperior',
             0x5E => 'sixinferior',
             0x5F => 'underscoredbl',   #?
             0x61 => 'fi',
             0x62 => 'fl',
             0x63 => 'ff',
             0x64 => 'ft',
             0x65 => 'ffi',
             0x66 => 'ffl',
             0x67 => 'fff',
             0x68 => 'fft',
             0x69 => 'ij',
             0x6A => 'dotlessj',
             0x6B => 'oneoneth',
             0x6C => 'onehalf',
             0x6D => 'onethird',
             0x6E => 'twothirds',
             0x6F => 'onequarter',
             0x70 => 'threequarters',
             0x71 => 'onefifth',
             0x72 => 'twofifths',
             0x73 => 'threefifths',
             0x74 => 'fourfifths',
             0x75 => 'onesixth',
             0x76 => 'fivesixths',
             0x77 => 'oneeighth',
             0x78 => 'threeeights',
             0x79 => 'fiveeights',
             0x7A => 'seveneights',
             0x7B => 'fractionbar',
             0x7C => 'brokenbar',
             0x9A => 'trademark',
           # 0xA9 => 'copyright',
           # 0xAE => 'registered',
             );
    } else {                    # edition earlier than 3
        %expertVec = 
            (  0041 => "copyright",
               0042 => "registered",
               0043 => "trademark",
               0044 => "dollaroldstyle",
               0045 => "brokenbar",
               0046 => "periodcentered",
               0047 => "asciigrave",      # ??
               
               0050 => "peseta",              # ??
               0051 => "pesetatwo",             # peseta?
               0052 => "pesetathree",
               0053 => "lira",
               0055 => "aovers",
               0056 => "fractionalt",
               0057 => "onealt",          # ??

               0072 => "ff",
               0073 => "fi",
               0074 => "fl",
               0075 => "ft",
               0076 => "fff",
               0077 => "ffi",

               0100 => "ffl",
               0101 => "fft",
               0102 => "zerosuperior",
               0103 => "onesuperior",
               0104 => "twosuperior",
               0105 => "threesuperior",
               0106 => "foursuperior",
               0107 => "fivesuperior",

               0110 => "sixsuperior",
               0111 => "sevensuperior",
               0112 => "eightsuperior",
               0113 => "ninesuperior",
               0114 => "parenleftsuperior",
               0115 => "parenrightsuperior",
               0116 => "dotcentered",
               0117 => "commacentered",

               0120 => "dollarsuperior",
               0121 => "centsuperior",
               0122 => "zeroinferior",
               0123 => "oneinferior",
               0124 => "twoinferior",
               0125 => "threeinferior",
               0126 => "fourinferior",
               0127 => "fiveinferior",

               0130 => "sixinferior",
               0131 => "seveninferior",
               0132 => "eightinferior",
               0133 => "nineinferior",
               0134 => "Eng",
               0135 => "eng",
               0136 => "underscoredbl",

               0140 => "capitalgrave",
               0141 => "asuperior",
               0142 => "esuperior",
               0143 => "isuperior",
               0144 => "lsuperior",
               0145 => "msuperior",
               0146 => "nsuperior",
               0147 => "osuperior",

               0150 => "rsuperior",
               0151 => "ssuperior",
               0152 => "tsuperior",
               0153 => "hyphensuperior",  # ??
               0154 => "oneoneth",        # ??
               0155 => "onehalf",
               0156 => "onethird",
               0157 => "twothirds",

               0160 => "onequarter",
               0161 => "threequarters",
               0162 => "onefifth",
               0163 => "twofifths",
               0164 => "threefifths",
               0165 => "fourfifths",
               0166 => "onesixth",
               0167 => "fivesixths",

               0170 => "oneeighth",
               0171 => "threeeights",
               0172 => "fiveeights",
               0173 => "seveneights",
               0174 => "fraction",
               0175 => "dotlessj",
               0176 => "ij",

               0242 => "centoldstyle",
               0243 => "kelvin",          # weird looking k ?
               0244 => "capitalbreve",

               0250 => "asciibreve",      # ?
               0251 => "twoperiods",      # short ellipses?
               0252 => "capitalacute",
               0254 => "capitaltilde",
               0255 => "asciitilde",
               0256 => "capitalhungarumlaut",
               0257 => "asciihungarumlaut", # ?

               0261 => "capitalring",
             # 0263 => "cedilla",
             # 0264 => "cedilla",         # what's the difference?
               0271 => "ogonek",
               0272 => "asciiacute",
               
               0301 => "capitalmacron",
               0303 => "fiveeightssmall",
               0304 => "seveneightssmall",
               0320 => "asciiring",       # ?
               0352 => "capitaldotaccent",
               0365 => "threeeightssmall",
               0372 => "asciidotaccent",
               );
    }

    open (AFM, "<$afm") or die "could not open $afm";
    open (NEW, ">$tmp") or die "cannot write to $tmp";
    
    while (<AFM>) {
        # Sanity check: look at the name to see if this is really
        # an expert font.
        if (/^FontName\s+(\S+)/) {
            my $name = $1;
            if ($name =~ /Ex/) {
                print "$tag: renaming $name\n";
            }
            else {
                close NEW;
                close AFM;
                unlink $tmp;
                return;
            }
        }
             
        # Match the EncodingScheme, so we can change it.
        if (/^EncodingScheme\s+\S+/) {
            $_ = "EncodingScheme FontSpecific\n";
        }
        
        # Match character lines such as the following:
        #   C 36 ; WX 467 ; N dollar ; B 29 -74 404 735 ;
        # Bind $1 to character number, $2 to name.

        if (/^C\s+(\d+)\s*;.*N\s+(\w+)/) {
            # See if that number is in our new encoding vector
            if ($expertVec{$1}) {
                my $new = $expertVec{$1};
                # If the name is the same, we can abort,
                # and also note that the experts need no reencoding.
                if ($new eq $2) {
                    print "$tag: rename aborted, $new eq $2\n";
                    $metrics{$fd}{expertokay} = 1;
                    close(NEW);
                    close(AFM);
                    return;
                }
                $_ =~ s/N\s+(\w+)/N $new/;
            }
        }
        print NEW "$_";
    }
    close (NEW);
    close (AFM);
    rename ($tmp, $afm) or die "could not rename $tmp to $afm";
}

## Thanks to Randall L. Schwartz for this one.
## http://www.stonehenge.com/merlyn/UnixReview/col30.html
sub deep_copy {
    my $this = shift;
    if (not ref $this) {
        $this;
    } elsif (ref $this eq "ARRAY") {
        [map deep_copy($_), @$this];
    } elsif (ref $this eq "HASH") {
        +{map { $_ => deep_copy($this->{$_}) } keys %$this};
    } else { die "what type is $_?" }
}

# EOF
